use simcore_rs::{Time, SimContext, Facility, RandomVar, simulation};
use rand::{distributions::Uniform, rngs::SmallRng, SeedableRng, Rng};

// helper constants
const STOP_TIME: Time = 480000.0;
const SEED_A   :  u64 = 100000;
const SEED_S   :  u64 = 200000;

fn main() {
	// pseudo random number generators referenced from within the main process
	let mut rng_a = SmallRng::seed_from_u64(SEED_A);
	let mut rng_s = SmallRng::seed_from_u64(SEED_S);
	
	// the lifetimes are automatically extended until the end of the scope
	let joe = &Facility::new();
	let rv = &RandomVar::new();
	
	// the main process
	simulation(|sim| async move {
		// activate a process to generate the customers
		sim.activate(async move {
			let dist = Uniform::new(12.0, 24.0);
			
			// wait some time before activating the first customer
			sim.advance(rng_a.sample(dist)).await;
			
			// generate new customers until the store closes officially
			while sim.now() < STOP_TIME {
				// activate the next customer
				sim.activate(Customer {
					joe, rv, rng: SmallRng::from_seed(rng_s.gen())
				}.actions(sim));
				
				// wait some time before activating the next customer
				sim.advance(rng_a.sample(dist)).await;
			}
		});
		
		// wait until the store closes
		sim.advance(STOP_TIME).await;
		
		// finish processing the queue (no more customers arrive)
		joe.seize().await;
	});
	
	println!("Stats: {:.3}", rv);
}

/// Customer process with access to the barber and a random number generator.
struct Customer<'j> { joe: &'j Facility, rv: &'j RandomVar, rng: SmallRng }

impl Customer<'_> {
	pub async fn actions(mut self, sim: SimContext<'_>) {
		// access the barber and record the time for the report
		let time = sim.now();
		self.joe.seize().await;
		self.rv.tabulate(sim.now() - time);
		
		// spend time
		sim.advance(self.rng.gen_range(12.0, 18.0)).await;
		// release the barber
		self.joe.release();
	}
}
